# User Notifications

> ℹ️ Till August 2020, we used the old `NSUserNotificationCenter`, and all code related to it was written in Objective-C. August 2020, we decided to implement the `Smart Suggestions` feature, which was written in Swift and utilized features of the new `UNNotificationCenter` (for the Agent, not so new for the macOS).\
> We had a working model of presenter for both user notification centers presenter in our POC of the `Smart Suggestions` feature and decided to port it in the Agent.\
> At that moment, AFXSetappAgentRoot used two classes working with local & remote notification accordingly.\
> We decided to port those classes to Swift cause thought it was faster for releasing a new feature.

## APN

We utilize Production Apple Push Notifications ENV (`https://api.push.apple.com/3/device/<token>`) on production Setapp ENV, and Sandbox env (`https://api.sandbox.push.apple.com/3/device/<token>`) on other Setapp envs.

### AWS SNS & Certificates regenerating
We use Amazon Simple Notification Service to send push notifications from backend. AWS SNS can't send push notifications using universal (iOS + macOS) certificate with combined ENVs (Production + Sandbox).
For that reason when creating a new pair of certificates you should create two of them:
  * `Apple Push Notification service SSL (Sandbox)` and 
  * `macOS Apple Push Notification service SSL (Production)`

AWS requires .p12 certificate, so you'll need to import downloaded certificate form developer.apple.com into your keychain (if you created a certificate request) and export certificate with private key in p12 format. Password could be found in 1Password, or you can change it as you wish, just tell the new password to the Site Reliability Engineers.

To check if everything is configured properly you can at first test exported certs locally (you can find instructions below under the `Payload samples` spoiler). After that you can upload certificate to the AWS SNS (Service Reliability Engineers can help) and check how everything works from the Setapp admin pannel of the corresponding ENV (prod or sandbox).

> 💁 Schedule a new reminder a month before a new certificate expires.

## How it works

### Intro
There are ~two~ three types of notifications:
* local user notifications,
* remote user notifications,
* remote background update notifications (aka **silent**).

Setapp uses `local user notifications` & `remote **silent** notification`.

> ℹ️ Differences between remote user notification & remote background update notifications:
> * remote user notification contains `alert`, `badge`, `sound` keys in the `aps` dictionary and displayed (sound is played, a badge is shown) by the macOS/iOS immediately on arrival.
> * **silent** variant has no those keys.
>
> You can read more about remote notifications payload [here](https://developer.apple.com/library/archive/documentation/NetworkingInternet/Conceptual/RemoteNotificationsPG/CreatingtheNotificationPayload.html#//apple_ref/doc/uid/TP40008194-CH10-SW1).

> ⚠️ Another point to tell is that `AFXRemoteNotificationsManager` doesn't care about remote notification type and can handle any remote notification. But once again: our backend sends only **silent** notifications.

### AFXSetappAgentRoot
A delegate of both old `NSUserNotificationCenter` & new `UNUserNotificationCenter`. This is where user notification handling starts.
A delegate of both Local & Remote notification managers. Thus can provide a lot of helpful info for those managers, especially for the Remote one.\

⚠️ Any user notification manager should report to it's delegate (`AFXSetappAgentRoot` 😉) start & finish events of the **local** user notification processing after used clicked on the notification or pressed an action button on it.\
The reason is simple: there is a logic in `AFXSetappAgentRoot` that will terminate the Agent after processing a local user notification if it was launched when the user clicked on the notification and not as a LaunchAgent. If Agent was launched from the notification, it will act as a client and will launch another correct instance of itself as a LaunchAgent.\
> ℹ️ Many of our **silent** notifications, despite being remote, become local.

#### Local notifications
The code for handling local user notifications in `AFXSetappAgentRoot` is pretty self-explanatory.

You can find more in `AFXLocalNotificationsManager` class.

#### Remote notifications (**silent**)
Since we use **silent** remote notifications, the `userNotificationCenter:willPresentNotification:withCompletionHandler` is a wrong place to start looking for remote notifications handling. You should start looking from the `application:didReceiveRemoteNotification:`.

Coming silently from our back-end, the notification `userInfo` dictionary (which is the only thing we have, truthfully speaking) contain all the info we need to proceed: remote notification type, actions the application needs to take, payload with back-end customer id, notification id on the back-end and a flag if the notification banner should be displayed.

You can find more in `AFXRemoteNotificationsManager` class.

## Objects

### AFXUserNotification
This class and its nested enums & structs are designed to create a new `UNNotificationRequest` or old `NSUserNotification` since we can't get rid of the old `NSUserNotificationCenter`.

#### `userInfo` & `payload` dictionaries
We heavily utilize the `userInfo` dictionary & even have another dictionary under the `payload` key.\
There are 2 enums `AFXUserNotification.UserInfoKey` & `AFXUserNotification.PayloadKey` describing all known & used keys in those dictionaries. And there is extension for `Dictionary` with helper functions to get a new dictionary with typed & untyped keys.\
Additionally, there are extensions for dictionaries (user info & payload) that provide easy & typed☝️ access to the properties. 
And there are even type-aliases to shorten dictionary type description.

##### `userInfo`
The heavily used dictionary by local notifications (there are some plans to move local notification properties to the `payload` dictionary).
It has a lot of possible keys, but the main things are local & remote notifications types, payload & remote notification actions.

##### `payload`
It is utilized by remote notifications sent by our back-end.
It contains a back-end user identifier it was addressed to, a back-end notification identifier, and a flag if the notification should be displayed or not.

#### Category, Action, Attachments
Just simple objects to create a corresponding object of the new `UNNotificationCenter`.

#### LocalIdentifier
It is used to describe all known identifiers of user notifications. It has an identifier string builder `localIdentifier`. That string is used to find similar notifications in the list of delivered notifications and delete them before presenting a new notification with the same identifier.

### AFXUserNotificationsPresenter
It presents user notifications in the User Notification Center.

Utilizes one of the underlying classes `AFXUserNotificationsPresenter.Old` (to deliver local user notifications uses the old `NSUserNotificationCenter`) or `AFXUserNotificationsPresenter.New` (to deliver local user notifications uses the new `UNUserNotificationCenter`) depending of available OS version.\

This set of objects is pretty straightforward. And you can read documentation comments in code.

### AFXLocalNotificationsManager
A class to build, present & handle local notifications such as application updates, new applications in Setapp etc.

You can find heavy utilization of the `userInfo` dictionary for building & handling local user notifications.\
The most tricky parts are using URLs of files with NSKeyedArchived set of the `AFXRunningApplication` to show & handle user notification about available updates for vendor application updates. Notification about update results is worth mentioning too.\
It is worth mentioning that we should tell the delegate when we started & finished handling the user notification.

### AFXRemoteNotificationsManager
Handles **silent** remote notifications & presents them if needed.

`userInfo` dictionary can contain an array of actions Setapp application needs to take after receiving the notification.

Both types contain a payload dictionary containing the back-end customer identifier, back-end notification identifier, and if the local notification banner should be presented.

Setapp application should execute actions from the `userInfo` dictionary for any remote notification type. And there are only two types: `.service` & `.notification`.

#### RemoteNotificationType.Service
Really silent notification. Just execute actions.

#### RemoteNotificationType.Notification
This kind of notifications requires that we should fetch additional info from the back-end (like title, body & even actions for the notification) and present it if needed the notification as **local**.
> We'll have to notify the delegate about processing opened by the user notification.

If the notification was shown we should mark the remote user notification as read & execute additional actions connected with button identifier and stored in the `AFXNotification`. For more details, you can look can start looking into the `func handleOpenAFXNotification(_:)` ad dive deeper.

## Updating certificates
Just the usual. Our back-end team has access to update certificates on AWS SNS service.\
Next time it would be great to generate iOS Production/Development certificate & select iOS certificate type one on SNS.\
Reason: it works, and we will have one certificate for development & production environments.

After updating, setup a reminder in Slack to the `#setapp-cocoa-devs` channel at least 15 days before certificate expiration.\
Sample: `/remind #setapp-cocoa-devs Generate new production certificates for push notifications on the Apple Developer portal & share it with the backend-team and tell them to update it on AWS SNS service at 12:00pm 15 sept 2022` 

And don't forget to update certificates and private keys in [1Password note `macOS.APNS`](https://start.1password.com/open/i?a=BIKTJE3ASJCA3NX5XWVZCO6PIY&v=qijy3ae65r6kor4rdwojxrbdjm&i=nbm252ez4zfz7lkyuga7saop6m&h=macpaw.1password.com).

## Testing
Pretty sure it will be easy to build local notifications (except for maybe the vendor application updates with running applications). 

#### w/ backend
You can send a limitless amount of remote notifications from the backend using our [admin site](https://admin.setapp.com/customer_notification/create).

#### w/o backend (almost)
You will need:
* A JSON payload from the back-end (make a breakpoint in the `application:didReceiveRemoteNotification:`, ask for the specific push and dump the dictionary into the JSON string, or write it by yourself).
* Device push notification token (we have it in the UserDafaults `defaults read com.setapp.DesktopClient APNSDeviceTokenString`).
* Development or Production `.p12` certificate with private key depends on what environment (production/sandbox) we will test. And you will need a password for the exported private key. You can find certificates in 1Password. 

You can send push notifications with [NWPusher](https://github.com/noodlewerk/NWPusher) or with `cURL`.

<details><summary>Payload samples</summary>
<details><summary>Silent service payload</summary>

```json
{
  "payload": {
    "customerId": 1547165,
  },
  "type": "service",
  "actions": [
    "refetch-customer",
    "refetch-devices",
  ],
}
```

</details>

<details><summary>Silent notification w/ local alert</summary>

```json
{
  "payload": {
    "notificationId": 6072797,
    "customerId": 1547165,
    "showBanner": true,
  },
  "type": "new-notification",
  "actions": [
    "refetch-customer",
    "refetch-devices",
  ],
}
```
</details>

<details><summary>cURL samples</summary>

You will need a Homebrew formula `curl-openssl` (don't forget to add it to the PATH in your zsh profile `export PATH="/usr/local/opt/curl-openssl/bin:$PATH"`).

Template:
```sh
curl -v \
  --http2 \
  --cert-type P12 --cert <p12_certificate_path>:<p12_certificate_password> \
  -H "apns-topic: com.setapp.DesktopClient.SetappAgent" \
  -H "apns-priority: 10" \
  -H "apns-push-type: <push_type>" \
  -d '<your payload here>' \
  https://api.push.apple.com/3/device/<device_token>
```

Silent service notification sample:
```sh
curl -v \
  --http2 \
  --cert-type P12 --cert <p12_certificate_path>:<p12_certificate_password> \
  -H "apns-topic: com.setapp.DesktopClient.SetappAgent" \
  -H "apns-push-type: alert" \
  -H "apns-priority: 10" \
  -d '{
    "payload": {
      "customerId": 1547165,
    },
    "type": "service",
    "actions": [
        "refetch-customer",
        "refetch-devices",
    ],
  }' \
  https://api.push.apple.com/3/device/43aa77ac4a2d26fcaf105e8c5ca890fbaf6a79ebb22be08e46616072a08cc712
```

Silent notification w/ local alert sample:
```sh
curl -v \
  --http2 \
  --cert-type P12 --cert <p12_certificate_path>:<p12_certificate_password> \
  -H "apns-topic: com.setapp.DesktopClient.SetappAgent" \
  -H "apns-push-type: alert" \
  -H "apns-priority: 10" \
  -d '{
    "payload": {
      "notificationId": 6072797,
      "customerId": 1547165,
      "showBanner": true,
    },
    "type": "new-notification",
    "actions": [
        "refetch-customer",
        "refetch-devices",
    ],
  }' \
  https://api.push.apple.com/3/device/43aa77ac4a2d26fcaf105e8c5ca890fbaf6a79ebb22be08e46616072a08cc712
```

> `apns-push-type: alert` because `background` type requires `apns-priority: 5` and I never managed to receive them.
</details>

<details><summary>Payload test</summary>
When you generate new cert you will want to test if app receives notifications.

For that reason you can add a custom alert payload that will be displayed. For that just add `"aps"` to your payload.
```json
"aps": {
  "alert": {
    "title": "Hi there",
    "body": "Will re-fetch customer",
  }
}
```

You can mix up it with the Silent upload.

Sample:
```sh
curl -v \                                                                                             ● develop 
  --http2 \
  --cert-type P12 --cert <p12_certificate_path>:<p12_certificate_password> \
  -H "apns-topic: com.setapp.DesktopClient.SetappAgent" \
  -H "apns-push-type: alert" \
  -H "apns-priority: 10" \
  -d '{
    "payload": {
      "customerId": 1547165,
    },
    "type": "service",
    "actions": [
        "refetch-customer",
        "refetch-devices",
    ],
    "aps": {
      "alert": {
        "body": "Will refetch customer",
        "title": "Hi there",
      }
    }
  }' \
  https://api.sandbox.push.apple.com/3/device/c24fc90e423034790c3d400fc3f5612e4d1aa4fc11f7d603955e446eee611c38
```

</details>
</details>
